# Solitude is a interactive story with four endings.
# Copyright (C) 2012 Ryan Martin
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os.path
import sys
import json
from Queue import Queue

import pyglet
from pyglet.gl import *
from pyglet import sprite
from pyglet import image
from pyglet import resource

from Viewport import FixedResolutionViewport

import GameData
import Interface
from Character import Character
from Entity import Entity
from Stage import Stage

class Shrine(Stage):
    def __init__(self):
        width, height=self.load_map("shrine.json")
        Stage.__init__(self, width, height)
        self.background=sprite.Sprite(resource.image("shrine_background.png"), batch=self.batch, group=self.backgroup)
        #self.building=sprite.Sprite(resource.image("shrine_foreground.png"), batch=self.batch, group=self.ogroups[1])
        #self.building.visible=False
        self.player.load_walkcycle(resource.image("farm_walkcycle.png"))
        self.player.set_idle("down")
        self.player.x=16*32
        self.player.y=3*32
        self.set_group(self.player)
        self.update_camera()
        self.astraea=Character(resource.image("astraea.png"), batch=self.batch, group=self.ogroups[1])
        self.astraea.x=16*32
        self.astraea.y=3*32
        self.set_group(self.astraea)
        self.queue=Queue()
        tree_image=resource.image("tree.png")
        tree_image.anchor_x=tree_image.width/2
        self.trees=[]
        for i,j in self.tree_locations:
            tree=sprite.Sprite(tree_image, batch=self.batch, group=self.ogroups[1])
            tree.x=i*32+16
            tree.y=j*32
            self.set_group(tree)
            self.trees.append(tree)
        cherry_tree_image=resource.image("cherry_tree.png")
        cherry_tree_image.anchor_x=cherry_tree_image.width/2
        self.cherry_trees=[]
        for i,j in self.cherry_tree_locations:
            cherry_tree=sprite.Sprite(cherry_tree_image, batch=self.batch, group=self.ogroups[1])
            cherry_tree.x=i*32+16
            cherry_tree.y=j*32
            self.set_group(cherry_tree)
            self.cherry_trees.append(cherry_tree)
        lantern_image=resource.image("lantern.png")
        lantern_image.anchor_x=lantern_image.width/2
        self.lanterns=[]
        for i,j in self.lantern_locations:
            lantern=sprite.Sprite(lantern_image, batch=self.batch, group=self.ogroups[1])
            lantern.x=i*32+16
            lantern.y=j*32
            self.set_group(lantern)
            self.lanterns.append(lantern)
        self.gate=sprite.Sprite(resource.image("torii.png"), batch=self.batch, group=self.ogroups[1])
        self.gate.x=31*32
        self.gate.y=49*32
        self.set_group(self.gate)
        deer_image=image.TextureGrid(image.ImageGrid(resource.image("deer.png"), 1, 4))
        self.deer=[]
        for i,j in self.deer_locations:
            deer=sprite.Sprite(deer_image[(0,2)], batch=self.batch, group=self.ogroups[1])
            deer.x=i*32
            deer.y=j*32
            self.set_group(deer)
            self.deer.append(deer)
        for i,j in self.deer_tile_locations:
            self.set_tile_action(i, j, self.toggle_deer)
        for i,j in self.rock_tile_locations:
            self.set_tile_action(i, j, self.toggle_rocks)
        for i,j in self.cherry_tile_locations:
            self.set_tile_action(i, j, self.toggle_cherries)
        self.seen_cherries=self.seen_deer=self.seen_rocks=False
        self.set_tile_action(15, 50, self.check_seen)
        self.set_tile_action(16, 50, self.check_seen)
        self.set_tile_action(15, 51, self.enter_shrine)
        self.set_tile_action(16, 51, self.enter_shrine)
        self.is_dream=True
        if Interface.music:
            self.music=resource.media("shrine.ogg")

    def toggle_deer(self):
        if not self.seen_deer:
            self.show_dialog("I wonder if we can feed them.")
            self.seen_deer=True

    def toggle_rocks(self):
        if not self.seen_rocks:
            self.show_dialog("They say you can't see all six rocks at the same time.")
            self.seen_rocks=True

    def toggle_cherries(self):
        if not self.seen_cherries:
            self.show_dialog("My friend said it looks like snow when the blossoms drop.")
            self.seen_cherries=True

    def check_seen(self):
        if not self.seen_deer:
            self.show_dialog("I want to see the deer first.")
        elif not self.seen_rocks:
            self.show_dialog("Can we visit the rock garden next?")
        elif not self.seen_cherries:
            self.show_dialog("Let's go see the cherry trees in bloom.")
        else:
            self.unset_collision(15*32, 51*32)
            self.unset_collision(16*32, 51*32)
            self.show_dialog("I can't wait to see what's inside.")

    def enter_shrine(self):
        self.toggle_shake()
        self.show_dialog("It's an earthquake!", self.end_stage)

    def end_stage(self):
        self.fade_out(lambda text=["The shrine stood for 900 years.", "Why couldn't it have stood for one more?"]:
                            self.show_dialog(text, lambda: Interface.switch_stage("Credits")))

    def load_map(self, mapname):
        width, height=Stage.load_map(self, mapname)
        for layer in self.layers:
            if layer["name"]=="Trees":
                self.tree_locations=self.get_locations(layer['data'], width, height)
            if layer["name"]=="Cherry Trees":
                self.cherry_tree_locations=self.get_locations(layer['data'], width, height)
            if layer["name"]=="Lanterns":
                self.lantern_locations=self.get_locations(layer['data'], width, height)
            if layer["name"]=="Deer":
                self.deer_locations=self.get_locations(layer['data'], width, height)
            if layer["name"]=="Deer Tiles":
                self.deer_tile_locations=self.get_locations(layer['data'], width, height)
            if layer["name"]=="Rock Tiles":
                self.rock_tile_locations=self.get_locations(layer['data'], width, height)
            if layer["name"]=="Cherry Tiles":
                self.cherry_tile_locations=self.get_locations(layer['data'], width, height)
        return width, height

    def update(self, dt):
        x=self.player.x
        y=self.player.y
        direction=self.player.direction
        Stage.update(self, dt)
        if x!=self.player.x or y!=self.player.y:
            self.queue.put((x,y,direction))
            if self.queue.qsize()<5:
                return
            x,y,direction=self.queue.get()
            self.astraea.x=x
            self.astraea.y=y
            if direction!=self.astraea.direction:
                self.astraea.set_walkcycle(direction)
            self.set_group(self.astraea)
        else:
            if self.astraea.image!=self.astraea.idle[direction]:
                self.astraea.set_idle(direction)


    #def show_building(self):
    #    self.building.visible=True

    #def hide_building(self):
    #    self.building.visible=False
