# Solitude is a interactive story with four endings.
# Copyright (C) 2012 Ryan Martin
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os.path
import sys
import json
from time import time
from math import sin, cos

import pyglet
from pyglet.gl import *
from pyglet import sprite
from pyglet import image
from pyglet import resource
from Viewport import FixedResolutionViewport

import GameData
import Interface
from Character import Character
from Entity import Entity

class Camera(object):
    def __init__(self):
        self.x=0
        self.y=0

class Stage(object):
    def __init__(self, width, height):
        self.window=Interface.window
        self.layers=[]
        self.layer_dict={}
        self.camera=Camera()
        self.shake_camera=False
        if GameData.options.fullscreen:
            self.viewport=FixedResolutionViewport(self.window, GameData.options.width, GameData.options.height, filtered=True)
        self.schedule_items=[]
        self.batch=pyglet.graphics.Batch()
        backdrop=pyglet.graphics.OrderedGroup(0)
        self.backdrop=sprite.Sprite(image.SolidColorImagePattern((0,0,0,255)).create_image(self.window.width, self.window.height),
                                    batch=self.batch, group=backdrop)
        self.backgroup=pyglet.graphics.OrderedGroup(1)
        self.ogroups=[pyglet.graphics.OrderedGroup(i) for i in xrange(100,(height*2+1)+101)]
        self.foregroup=pyglet.graphics.OrderedGroup(9999)
        self.dialog=pyglet.sprite.Sprite(pyglet.resource.image("dialog_box.png"))
        self.dialog.x=(GameData.options.width-self.dialog.image.width)/2
        self.dialog.y=24
        m=10
        self.dialog_text=label = pyglet.text.Label('', font_name='Arial', font_size=18, bold=True,
                          x=self.dialog.x+m, y=self.dialog.y+self.dialog.image.height-m, width=self.dialog.image.width-2*m, 
                          height=self.dialog.image.height-2*m, anchor_y='top', multiline=True)
        self.center_dialog=pyglet.sprite.Sprite(pyglet.resource.image("dialog_box.png"))
        self.center_dialog.x=(GameData.options.width-self.dialog.image.width)/2
        self.center_dialog.y=(GameData.options.height-self.dialog.image.height)/2
        m=10
        self.center_dialog_text=label = pyglet.text.Label('', font_name='Arial', font_size=18, bold=True,
                          x=self.center_dialog.x+m, y=self.center_dialog.y+self.dialog.image.height-m, width=self.center_dialog.image.width-2*m, 
                          height=self.center_dialog.image.height-2*m, anchor_y='top', multiline=True)
        self.yes_no=sprite.Sprite(resource.image("yes_no_dialog.png"))
        yes_no_offset=128
        self.yes_no.x=(GameData.options.width-self.yes_no.width)/2
        self.yes_no.y=(GameData.options.height-self.yes_no.height)/2-yes_no_offset
        self.caret=sprite.Sprite(resource.image("caret.png"))
        self.caret_locations={'yes':(14,76), 'no':(14,44)}
        self.caret.yes=False
        self.toggle_caret()
        self.action_map={}
        self.tile_actions={}
        self.keys=pyglet.window.key.KeyStateHandler()
        self.player=Character(resource.image("male_walkcycle.png"), batch=self.batch, group=self.ogroups[1])
        self.player.nomove=False
        self.fade=sprite.Sprite(image.SolidColorImagePattern((0,0,0,255)).create_image(self.window.width, self.window.height))
        self.fade.opacity=0
        self.is_dream=False
        self.dream=sprite.Sprite(resource.image("dream.png"))
        self.music=None
        self.dialog_draw=None
        self.fade_draw=None

    def load_walkcycle(self, im):
        walkcyclegrid=image.TextureGrid(image.ImageGrid(im, 4, 1))
        walkcycle={}
        idle={}
        duration=0.1
        for i, d in enumerate(["right", "down", "left", "up"]):
            cycle=image.ImageGrid(walkcyclegrid[i], 1, 9)
            for frame in cycle:
                frame.anchor_x=frame.width/2
            walkcycle[d]=image.Animation(map(lambda f: image.AnimationFrame(f, duration), cycle[1:]))
            idle[d]=cycle[0]
        return walkcycle, idle

    def new_sprite(self, i, j, im):
        s=sprite.Sprite(im, batch=self.batch, group=self.ogroups[1])
        s.x=i*32
        s.y=j*32
        self.set_group(s)
        return s

    def set_tile_action(self, i, j, action):
        self.tile_actions[(j,i)]=action

    def unset_tile_action(self, i, j):
        self.tile_actions[(j,i)]=lambda: False

    def fade_out(self, action=None):
        self.fade.opacity=0
        self.fade_draw=self.fade.draw
        def fade(dt, action=action):
            df=255
            if self.fade.opacity<255:
                self.fade.opacity+=df*dt
                if self.fade.opacity>255:
                    self.fade.opacity=255
            else:
                pyglet.clock.unschedule(fade)
                if action:
                    action()
        pyglet.clock.schedule(fade)

    def fade_in(self, action=None):
        self.fade.opacity=255
        self.fade_draw=self.fade.draw
        def fade(dt, action=action):
            df=255
            if self.fade.opacity>0:
                self.fade.opacity-=df*dt
                if self.fade.opacity<0:
                    self.fade.opacity=0
            else:
                pyglet.clock.unschedule(fade)
                self.fade_draw=None
                if action:
                    action()
        pyglet.clock.schedule(fade)
    
    def toggle_caret(self):
        if self.caret.yes:
            self.caret.yes=False
            self.caret.x=self.caret_locations['no'][0]+self.yes_no.x
            self.caret.y=self.caret_locations['no'][1]+self.yes_no.y
        else:
            self.caret.yes=True
            self.caret.x=self.caret_locations['yes'][0]+self.yes_no.x
            self.caret.y=self.caret_locations['yes'][1]+self.yes_no.y

    def get_locations(self, layer_data, width, height):
        locations=[]
        for i in xrange(0,width):
            for j in xrange(0, height):
                if layer_data[i+j*width]:
                    locations.append((i,height-1-j))
        return locations

    def load_map(self, mapname):
        layers=json.load(resource.file(mapname))['layers']
        for layer in layers:
            if layer['name']=='Collision':
                height=layer['height']
                width=layer['width']
                self.collision_map=[layer['data'][j*width:j*width+width] for j in xrange(height-1, -1, -1)]
        self.layers=layers
        return width, height

    def show_yes_no_dialog(self, text, action):
        self.yes_no.caret=True
        self.center_dialog_text.text=text
        self.player.nomove=True
        if not self.caret.yes:
            self.toggle_caret()
        def on_draw():
            self.center_dialog.draw()
            self.center_dialog_text.draw()
            self.yes_no.draw()
            self.caret.draw()
        def on_key_press(symbol, modifiers):
            key=pyglet.window.key
            if symbol==key.UP or symbol==key.DOWN:
                self.toggle_caret()
            elif symbol==key.SPACE:
                self.hide_dialog()
                if self.caret.yes:
                    action()
            return True
        self.dialog_draw=on_draw
        self.window.push_handlers(on_key_press)   

    def show_dialog(self, text='', action=None):
        self.player.nomove=True
        if text:
            if isinstance(text, list):
                self.d_texts=iter(text)
                self.dialog_text.text=self.d_texts.next()
            else:
                self.d_texts=iter([text])
                self.dialog_text.text=self.d_texts.next()
        def on_draw():
            self.dialog.draw()
            self.dialog_text.draw()
        def on_key_press(symbol, modifiers, action=action):
            key=pyglet.window.key
            if symbol==key.SPACE:
                try:
                    self.dialog_text.text=self.d_texts.next()
                except StopIteration:
                    self.hide_dialog()
                    if action:
                        action()
            return True
        self.dialog_draw=on_draw
        self.window.push_handlers(on_key_press) 

    def hide_dialog(self):
        self.dialog_draw=None
        self.player.nomove=False
        self.window.pop_handlers()

    def set_action(self, x, y, action):
        i=int(x/32)
        j=int(y/32)
        self.action_map[(j,i)]=action

    def unset_action(self, x, y):
        i=int(x/32)
        j=int(y/32)
        self.action_map[(j,i)]=lambda: False

    def set_collision(self, x, y):
        i=int(x/32)
        j=int(y/32)
        self.collision_map[j][i]=1

    def unset_collision(self, x, y):
        i=int(x/32)
        j=int(y/32)
        self.collision_map[j][i]=0

    def add_entity(self, entity):
        self.set_action(entity.x, entity.y, entity.action)
        self.set_collision(entity.x, entity.y)

    def collide(self, x, y):
        i=int(x/32)
        j=int(y/32)
        return self.collision_map[j][i]

    def set_group(self, s):
        if s==self.player:
            s.group=self.ogroups[(self.background.height/32-int((s.y-16)/32))*2-1]
        else:
            s.group=self.ogroups[(self.background.height/32-int((s.y)/32))*2]

    def action(self):
        i=int(self.player.x/32)
        j=int(self.player.y/32)
        if self.player.direction=="left":
            i-=1
        elif self.player.direction=="right":
            i+=1
        elif self.player.direction=="up":
            j+=1
        elif self.player.direction=="down":
            j-=1
        if (j,i) in self.action_map:
            self.action_map[(j,i)]()
        else:
            self.default_action()

    def default_action(self):
        pass

    def clear_handler_stack(self):
        try:
            while True:
                self.window.pop_handlers()
        except AssertionError:
            #Stack empty
            pass

    def on_draw(self):
        pyglet.gl.glLoadIdentity()
        pyglet.gl.glTranslatef(-self.camera.x, -self.camera.y, 0)
        self.batch.draw()
        pyglet.gl.glTranslatef(self.camera.x, self.camera.y, 0)
        if self.fade_draw:
            self.fade_draw()
        if self.dialog_draw:
            self.dialog_draw()
        if self.is_dream:
            self.dream.draw()

    def on_draw_begin(self):
        self.viewport.begin()

    def on_draw_end(self):
        self.viewport.end()

    def on_key_press(self, symbol, modifiers):
        key=pyglet.window.key
        if symbol==key.LEFT:
            self.player.set_walkcycle("left")
        elif symbol==key.RIGHT:
            self.player.set_walkcycle("right")
        elif symbol==key.UP:
            self.player.set_walkcycle("up")
        elif symbol==key.DOWN:
            self.player.set_walkcycle("down")
        elif symbol==key.SPACE:
            self.action()
        return False

    def on_key_release(self, symbol, modifiers):
        key=pyglet.window.key
        for sym in [key.LEFT, key.RIGHT, key.UP, key.DOWN]:
            if self.keys[sym]:
                self.on_key_press(sym, modifiers)
                return False
        if symbol==key.LEFT:
            self.player.set_idle("left")
        elif symbol==key.RIGHT:
            self.player.set_idle("right")
        elif symbol==key.UP:
            self.player.set_idle("up")
        elif symbol==key.DOWN:
            self.player.set_idle("down")
        return False

    def update(self, dt):
        if self.player.nomove:
            self.update_camera()
            return
        key=pyglet.window.key
        v=200
        dx=dy=0
        if self.keys[key.LEFT]:
            dx=-v*dt
        if self.keys[key.RIGHT]:
            dx=v*dt
        if self.keys[key.UP]:
            dy=v*dt
        if self.keys[key.DOWN]:
            dy=-v*dt
        new_x=self.player.x+int(dx)
        new_x2=self.player.x+int(dx/2)
        new_y=self.player.y+int(dy)
        new_y2=self.player.y+int(dy/2)
        if new_x>0 and new_x<self.background.width and new_y>0 and new_y<self.background.height:
            bit=self.collide(new_x, new_y)
            bit2=self.collide(new_x2, new_y2)
            if not bit and not bit2:
                i=int(self.player.x/32)
                j=int(self.player.y/32)
                self.player.x=new_x
                self.player.y=new_y
                i2=int(self.player.x/32)
                j2=int(self.player.y/32)
                self.update_camera()
                self.set_group(self.player)
                if i!=i2 or j!=j2:
                    if (j2,i2) in self.tile_actions:
                        self.tile_actions[(j2,i2)]()

    def toggle_shake(self):
        if self.shake_camera:
            self.shake_camera=False
        else:
            self.shake_camera=True
            self.base_time=time()
                

    def update_camera(self):
        self.camera.x=self.player.x-GameData.options.width/2
        self.camera.y=self.player.y-GameData.options.height/2
        if self.shake_camera:
            t=time()-self.base_time
            coef=10
            self.camera.x+=coef*sin(coef*t)
            self.camera.y+=coef*cos(coef*t)
        if self.camera.x<0:
            self.camera.x=0
        right_limit=self.background.width-GameData.options.width
        if self.camera.x>right_limit:
            self.camera.x=right_limit
        if self.camera.y<0:
            self.camera.y=0
        top_limit=self.background.height-GameData.options.height
        if self.camera.y>top_limit:
            self.camera.y=top_limit
        if self.background.width<GameData.options.width:
            self.camera.x=(self.background.width-GameData.options.width)/2
        if self.background.height<GameData.options.height:
            self.camera.y=(self.background.height-GameData.options.height)/2

    def display(self, music):
        self.clear_handler_stack()
        if GameData.options.fullscreen:
            self.window.push_handlers(on_draw=self.viewport.begin)
        self.window.push_handlers(self.on_draw, self.keys, self.on_key_press, self.on_key_release)
        self.window.push_handlers(self.keys)
        pyglet.clock.schedule_interval(self.update, 1/60.0)
        if music and self.music:
            self.music_player=pyglet.media.Player()
            self.music_player.queue(self.music)
            self.music_player.eos_action=self.music_player.EOS_LOOP
            self.music_player.play()
        if GameData.options.fullscreen:
            self.window.push_handlers(on_draw=self.viewport.end)
